<!DOCTYPE html>
<html lang="{{ site.lang | default: "en-US" }}">
<script type="text/javascript">
window.location.href = 'INTRO.html'
</script>
If you see this page Javascript may be disabled in your browser.  
Please add INTRO.html to the URL to see the documentation. 
</html>
